{-# LANGUAGE RankNTypes, MultiParamTypeClasses, NoMonomorphismRestriction, TypeOperators, FunctionalDependencies, TypeFamilies, FlexibleContexts, FlexibleInstances, DatatypeContexts, UndecidableInstances, ScopedTypeVariables #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  GHC.HetMet.IGArrow
-- Copyright   :  none
-- License     :  public domain
--
-- Maintainer  :  Adam Megacz <megacz@acm.org>
-- Stability   :  experimental
-- Portability :  portable

module GHC.HetMet.IGArrow (
  IGArrow(..),
  IGArrowDrop(..),
  IGArrowCopy(..),
  IGArrowSwap(..),
  IGArrowLoop(..),

  IGArrowEval(..),
  IGArrowConstant(..),
  IGArrowLiteral(..),

  IGArrowReify(..),
  IGArrowReflect(..),

  -- IGArrowSum(..),  ga_inl, ga_inr,
  -- IGArrowProd(..),

) where
import Control.Category
import GHC.HetMet.GArrow
import GHC.HetMet.Arrow
import GHC.HetMet.GArrowFullyEnriched
import Prelude          hiding (id, (.))





------------------------------------------------------------------------
-- Internal GArrows
--
-- | An "internal generalized arrow" is a GArrow except that it uses
-- some *other* GArrow in place of Haskell's function space.
--
class GArrow g (**) u => IGArrow g (**) u gg (***) uu | gg -> (***), (***) -> uu where
  iga_id        :: g u (gg x x)
  iga_comp      :: g ((gg x y) ** (gg y z)) (gg x z)
  iga_first     :: g (gg x y) (gg (x *** z) (y *** z))
  iga_second    :: g (gg x y) (gg (z *** x) (z *** y))
  iga_cancell   :: g u (gg (uu***x)         x)
  iga_cancelr   :: g u (gg    (x***uu)      x)
  iga_uncancell :: g u (gg     x      (uu***x))
  iga_uncancelr :: g u (gg     x         (x***uu))
  iga_assoc     :: g u (gg ((x*** y)***z ) ( x***(y ***z)))
  iga_unassoc   :: g u (gg ( x***(y ***z)) ((x*** y)***z ))

class IGArrow g (**) u gg (***) uu => IGArrowCopy g (**) u gg (***) uu where
  iga_copy      :: g u (gg x (x***x))

class IGArrow g (**) u gg (***) uu => IGArrowDrop g (**) u gg (***) uu where
  iga_drop      :: g u (gg x u)

class IGArrow g (**) u gg (***) uu => IGArrowSwap g (**) u gg (***) uu where
  iga_swap      :: g u (gg (x***y) (y***x))

class IGArrow g (**) u gg (***) uu => IGArrowLoop g (**) u gg (***) uu where
  iga_loopr    :: g (gg (x***z) (y***z)) (gg x y)
  iga_loopl    :: g (gg (z***x) (z***y)) (gg x y)

class IGArrow g (**) u gg (***) uu => IGArrowLiteral g (**) u gg (***) uu t r where
  iga_literal  :: g t (gg uu r)

class IGArrow g (**) u gg (***) uu => IGArrowEval g (**) u gg (***) uu r t where
  iga_eval      :: g (gg uu r) t

class IGArrow g (**) u gg (***) uu => IGArrowConstant g (**) u gg (***) uu t r where
  iga_constant  :: g t (gg uu r)

class IGArrow g (**) u gg (***) uu => IGArrowReify g (**) u gg (***) uu x y r q where
  iga_reify     :: g (g x y) (gg r q)

class IGArrow g (**) u gg (***) uu => IGArrowReflect g (**) u gg (***) uu r q x y where
  iga_reflect   :: g (gg r q) (g x y)



{-
-- | This is named "kappa" for its similarity to an operator in
--   Hasegawa's kappa-calculus, but there isn't actually a formal
--   connection.
class GArrow g (**) u => IKappa g (**) u where
  iga_kappa     :: forall a b c uu .
                   (forall gg (***) . (IGArrow     g (**) u gg (***) uu,
                                       IGArrowCopy g (**) u gg (***) uu,
                                       IGArrowSwap g (**) u gg (***) uu,
                                       IGArrowDrop g (**) u gg (***) uu) => g (gg uu a) (gg b c)) ->
                   (forall gg (***) . (IGArrow     g (**) u gg (***) uu,
                                       IGArrowCopy g (**) u gg (***) uu,
                                       IGArrowSwap g (**) u gg (***) uu,
                                       IGArrowDrop g (**) u gg (***) uu) => g u (gg (a***b) c))
  -- TO DO: probably need l/r versions of the above

--class IGArrow g (**) u gg (***) uu => IGArrowKappa g (**) u gg (***) uu where
--  iga_kappa     :: g (x**(gg uu a)) (gg b c) -> g x (gg (a***b) c)
  -- TO DO: probably need l/r versions of the above





------------------------------------------------------------------------------
-- Self-Internal GArrows

--
-- | A self-internal GArrow is, well, internal to itself
--
--class IGArrow g (**) u g (**) u => SelfInternalGArrow g (**) u

--
-- | Self-internal GArrows have curry/apply
--
-- instance SelfInternalGArrow g (**) u => GArrowApply g (**) u gg where
--   ga_applyl =    :: g (x**(g x y)   ) y
--   ga_applyr    :: g (   (g x y)**x) y
-- 
--
-- | Self-internal GArrows have curry/apply
--
-- instance SelfInternalGArrow g (**) u gg (***) uu => GArrowCurry g (**) u gg where
--   ga_curryl    :: g (x**y) z  ->  g x (g y z)
--   ga_curryr    :: g (x**y) z  ->  g y (g x z)





------------------------------------------------------------------------------
-- Instances

--
-- | Every GArrow is internal to the GArrow instance on (->)
--
{-
instance GArrow g (**) u => IGArrow (->) (,) () g (**) u where
  iga_id        _     = id
  iga_comp      (f,g) = f >>> g
  iga_first           = ga_first
  iga_second          = ga_second
  iga_cancell   _     = ga_cancell
  iga_cancelr   _     = ga_cancelr
  iga_uncancell _     = ga_uncancell
  iga_uncancelr _     = ga_uncancelr
  iga_assoc     _     = ga_assoc
  iga_unassoc   _     = ga_unassoc
instance GArrowCopy g (**) u => IGArrowCopy (->) (,) () g (**) u where
instance GArrowSwap g (**) u => IGArrowSwap (->) (,) () g (**) u where
instance GArrowDrop g (**) u => IGArrowDrop (->) (,) () g (**) u where
-}

data
    In g x y
    = In (g x y)
instance GArrow g (**) u     => IGArrow (->) (,) ()     (In g) (**) u where
instance GArrowCopy g (**) u => IGArrowCopy (->) (,) () (In g) (**) u where
instance GArrowSwap g (**) u => IGArrowSwap (->) (,) () (In g) (**) u where
instance GArrowDrop g (**) u => IGArrowDrop (->) (,) () (In g) (**) u where

------------------------------------------------------------------------
-- Externalization
--
-- | An IGArrow may be "externalized" to form a normal generalized
-- arrow.  If the IGArrow was an instance of class IGArrowXX, the
-- externalization will be an instance of GArrowYY.
--
-- TODO: I should be one level deeper here: assuming an (IGArrow
-- (IGArrow g)), create an (IGArrow g).
--

data
    Ext g x y
    = Ext (g x y)


--
-- | Every IGArrow of (->) is a GArrow
--
instance IGArrow (->) (,) () g (**) u => Category (Ext g) where
  id                = Ext (iga_id ())
  (Ext g) . (Ext f) = Ext (iga_comp (f,g))

instance IGArrow (->) (,) () g (**) u => GArrow (Ext g) (**) u where
--  ga_first     (Ext f) = Ext $ iga_first f
--  ga_second    (Ext f) = Ext $ iga_second f
  ga_cancell           = Ext $ iga_cancell ()
  ga_cancelr           = Ext $ iga_cancelr ()
  ga_uncancell         = Ext $ iga_uncancell ()
  ga_uncancelr         = Ext $ iga_uncancelr ()
  ga_assoc             = Ext $ iga_assoc ()
  ga_unassoc           = Ext $ iga_unassoc ()

instance IGArrowSwap (->) (,) () g (**) u => GArrowSwap (Ext g) (**) u where
instance IGArrowCopy (->) (,) () g (**) u => GArrowCopy (Ext g) (**) u where
instance IGArrowDrop (->) (,) () g (**) u => GArrowDrop (Ext g) (**) u where


--  iga_kappa     :: forall a b c uu . (forall gg (***) . IGArrow g (**) u gg (***) uu => g (gg uu a) (gg b c)) ->
--                   (forall gg (***) . IGArrow g (**) u gg (***) uu => g u (gg (a***b) c))
{-
    Could not deduce (GArrowDrop gg (***) uu,
                      GArrowSwap gg (***) uu,
                      GArrowCopy gg (***) uu)
      arising from a use of `homfunctor_inv'
    from the context (IGArrow (->) (,) () gg (***) uu)
-}

--homfunctor_inv :: forall a b c u .
--    (forall g (**) . (GArrowSwap g (**) u, GArrowCopy g (**) u, GArrowDrop g (**) u) => g u a -> g b c) ->
--    (forall g (**) . (GArrowSwap g (**) u, GArrowCopy g (**) u, GArrowDrop g (**) u) => g (a**b) c)

--
-- | Every GArrow is also an IKappa of (->) by virtue of GArrowFullyEnriched
--
instance IKappa (->) (,) () where
--  iga_kappa     :: forall a b c u .
--                   (forall g (**) . (IGArrow     (->) (,) () g (**) u,
--                                       IGArrowCopy (->) (,) () g (**) u,
--                                       IGArrowSwap (->) (,) () g (**) u,
--                                       IGArrowDrop (->) (,) () g (**) u) => (g u a) -> (g b c)) ->
--                   (forall g (**) . (IGArrow     (->) (,) () g (**) u,
--                                       IGArrowCopy (->) (,) () g (**) u,
--                                       IGArrowSwap (->) (,) () g (**) u,
--                                       IGArrowDrop (->) (,) () g (**) u) => () -> (g (a**b) c))
  iga_kappa (f :: forall g (**) .
                  (IGArrow (->) (,) () g (**) u,
                   IGArrowCopy (->) (,) () g (**) u,
                   IGArrowDrop (->) (,) () g (**) u,
                   IGArrowSwap (->) (,) () g (**) u) =>
                  (g u a)->(g b c)) ()
     = mork $ homfunctor_inv (\y -> case f (In y) of (In q) -> q)

mork :: forall u a b c . 
        (forall g (**) . (GArrowSwap g (**) u, GArrowCopy g (**) u, GArrowDrop g (**) u) => g (a**b) c)
 -> (forall g (**) . (IGArrow     (->) (,) () g (**) u,
                      IGArrowCopy (->) (,) () g (**) u,
                      IGArrowSwap (->) (,) () g (**) u,
                      IGArrowDrop (->) (,) () g (**) u) => (g (a**b) c))
mork = undefined
{-
         z :: forall g (**) . IGArrow (->) (,) () g (**) u => g (a**b) c
         z = let 
               qf :: forall g (**) . (GArrowSwap g (**) u, GArrowCopy g (**) u, GArrowDrop g (**) u) => g u a -> g b c
               qf y = f y
             in case homfunctor_inv qf of Ext q -> q
-}

{-
instance IGArrow (->) (,) () g (**) u => GArrowSTKC (Ext g)
-}

-- instance IGArrow g (**) u gg (***) uu => GArrowCopy g (**) u gg (***) uu where
--   iga_copy      = undefined
-- 
-- instance IGArrow g (**) u gg (***) uu => GArrowDrop g (**) u gg (***) uu where
--   iga_drop      = undefined
-- 
-- instance IGArrow g (**) u gg (***) uu => GArrowSwap g (**) u gg (***) uu where
--   iga_swap      = undefined
-- 
-- instance IGArrow g (**) u gg (***) uu => GArrowLoop g (**) u gg (***) uu where
--   iga_loopr    = undefined
--   iga_loopl    = undefined
-- 
-- instance IGArrow g (**) u gg (***) uu => GArrowLiteral g (**) u gg (***) uu t r where
--   iga_literal  = undefined
-- 
-- instance IGArrow g (**) u gg (***) uu => GArrowEval g (**) u gg (***) uu r t where
--   iga_eval      = undefined
-- 
-- instance IGArrow g (**) u gg (***) uu => GArrowConstant g (**) u gg (***) uu t r where
--   iga_constant  = undefined
-- 
-- instance IGArrow g (**) u gg (***) uu => GArrowReify g (**) u gg (***) uu x y r q where
--   iga_reify     = undefined
-- 
-- instance IGArrow g (**) u gg (***) uu => GArrowReflect g (**) u gg (***) uu r q x y where
--   iga_reflect   = undefined


-- | Haskell's function space is self-internal
-- instance SelfInternalGArrow (->) (,) ()

--
--     x , 1->a   |-    b ->c
--    --------------------------   [Kappa]
--     x          |- (a,b)->c
--
-- forall a b c x.
--   (forall g (**) u gg (***) uu .
--     IGArrow g (**) u gg (***) uu =>
--       g (x**(gg uu a)) (gg b c))
--   ->
--   (forall g (**) u gg (***) uu .
--     IGArrow g (**) u gg (***) uu =>
--       g x (gg (a***b) c))
-}